package com.lambelly.lambnes.platform.mappers;

import java.io.IOException;

import org.apache.log4j.Logger;

import com.lambelly.lambnes.LambNes;
import com.lambelly.lambnes.platform.cpu.NesCpuMemory;
import com.lambelly.lambnes.platform.ppu.NesPpuMemory;
import com.lambelly.lambnes.platform.interrupts.*;
import com.lambelly.lambnes.util.NumberConversionUtils;

public class MMC3Mapper extends Mapper0 implements Mapper
{
	private int[] prgRom = null; 
	private int[] chrRom = null;
	private int[] prgRam = null;

	private boolean addressInvert = false;
	private int prgRomSwapControl = 0;
	private int swapModeControl = 0;
	private int mirroringControl = 0;
	private int prgRomBankMode = 0;
	private boolean prgRamChipEnabled = false;
	private Logger logger = Logger.getLogger(MMC3Mapper.class);

	public static final int PRM_ROM_SWAP_8000_BFFF = 0;
    public static final int PRG_ROM_SWAP_A000_DFFF = 1;
    
    public static final int PRG_ROM_BANK_MODE_8000_9FFF_SWAPPABLE = 0;
    public static final int PRG_ROM_BANK_MODE_C000_DFFF_SWAPPABLE = 1;
    
    public static final int CONTROL_SWAP_2_1K_CHR_ROM_0000 = 0;
    public static final int CONTROL_SWAP_2_1K_CHR_ROM_0800 = 1;
    public static final int CONTROL_SWAP_1K_CHR_ROM_1000 = 2;
    public static final int CONTROL_SWAP_1K_CHR_ROM_1400 = 3;
    public static final int CONTROL_SWAP_1K_CHR_ROM_1800 = 4;
    public static final int CONTROL_SWAP_1K_CHR_ROM_1C00 = 5;
    public static final int CONTROL_SWAP_8K_PRG_ROM_8000 = 6;
    public static final int CONTROL_SWAP_8K_PRG_ROM_A000 = 7;
    
    public static final int MIRRORING_VERTICAL = 0;
    public static final int MIRRORING_HORIZONTAL = 1;
    
    public static final int PRG_RAM_WRITE_ENABLED = 0;
    public static final int PRG_RAM_WRITE_DISABLED = 1;
	
	public MMC3Mapper(NesCpuMemory cpuMemory, NesPpuMemory ppuMemory) 
	{
		this.setCpuMemory(cpuMemory);
		this.setPpuMemory(ppuMemory);
	}
	
	public void setMemoryFromHexAddress(int address, int value) throws IllegalStateException, IOException
	{
		// logger.info("write to address: " + address + " value: " + value);
		
		// determine whether odd or even 
		boolean addressEven = address % 2 == 0; 
		
		if (address >= 0x6000 && address <= 0x7FFF)
		{
			if (this.isPrgRamChipEnabled())
			{
				int prgRamIndex = address & 0x1FFF;
				// logger.info("setting prg ram address " + prgRamIndex + ": " + value);
				this.setPrgRamValue(prgRamIndex, value);
			}
			else
			{
				super.setMemoryFromHexAddress(address, value);
			}
		}
		else
		{
			// addresses in the range 0x8000 - 0xFFFF are not writable. They only affect the mapper state.
			if (address >= 0x0 && address <= 0x7FFF)
			{
				super.setMemoryFromHexAddress(address, value);
			}
		}
		
		if (address >= 0x2000 && address <= 0x3FFF)
		{
            if (this.getMirroringControl() == MMC3Mapper.MIRRORING_VERTICAL) 
            {  
            	this.getPpuMemory().establishMirroring(NesPpuMemory.MIRRORING_VERTICAL); 
            }
            else
            {
            	this.getPpuMemory().establishMirroring(NesPpuMemory.MIRRORING_HORIZONTAL); 
            }
		}
		if (address >= 0x4000 && address <= 0x5FFF)
		{
			// TODO
			logger.error("write to 0x4000 on MMC3, which isn't really implemented as of yet.");
		}
		if (address >= 0x6000 && address <= 0x6FFF)
		{
			// TODO
			logger.error("write to 0x6000 on MMC3, which isn't really implemented as of yet.");
		}
		else if (address >= 0x8000 && address <= 0x9FFF)
		{	
			if (addressEven)
			{
				//logger.info("write to 8000 register: " + value + " : " + NumberConversionUtils.generateBinaryStringWithleadingZeros(value, 8));
				// control register
				// CSxx xMMM
				this.setAddressInvert((value & 128) != 0);
				this.setPrgRomBankMode((value & 64) >> 6);
				this.setSwapModeControl(value & 7);
				
				// When S = 0, C000-DFFF will contain the second to last bank of ROM.
				// When S = 1, 8000-9FFF will contain the second to last bank of ROM.
				int banks = this.prgRom.length / LambNes.NUM_8_K;
				if (this.getPrgRomBankMode() == 1)
				{
					// 8000
					//logger.debug("swapping out 8000");
					System.arraycopy(this.getPrgRom(), ((banks - 2) * LambNes.NUM_8_K), this.getCpuMemory().getMemory(), 0x8000, LambNes.NUM_8_K);
				}
				else
				{
					// C000
					System.arraycopy(this.getPrgRom(), ((banks - 2) * LambNes.NUM_8_K), this.getCpuMemory().getMemory(), 0xC000, LambNes.NUM_8_K);
				}				
			}
			else
			{
				//logger.info("write to 8001 register: " + value + " : " + NumberConversionUtils.generateBinaryStringWithleadingZeros(value, 8) + " : chrRomSwapControl: " + this.swapModeControl() + " : PrgRomBankMode: " + this.getPrgRomBankMode());
				// data register
				switch(this.swapModeControl())
				{
					case MMC3Mapper.CONTROL_SWAP_2_1K_CHR_ROM_0000:
						logger.info("chr_rom 0000: address invert: " + this.isAddressInvert());
						if (!this.isAddressInvert())
						{
							// Select 2 1KB CHR bank at PPU $0000-$07FF 
							System.arraycopy(this.getChrRom(), (value * LambNes.NUM_1_K), this.getPpuMemory().getPatternTable0(), 0, LambNes.NUM_1_K);
							System.arraycopy(this.getChrRom(), ((value + 1) * LambNes.NUM_1_K), this.getPpuMemory().getPatternTable0(), 0x400, LambNes.NUM_1_K);
						}
						else
						{
							// Select 2 1KB CHR bank at PPU $1000-$17FF
							System.arraycopy(this.getChrRom(), (value * LambNes.NUM_1_K), this.getPpuMemory().getPatternTable1(), 0, LambNes.NUM_1_K);
							System.arraycopy(this.getChrRom(), ((value + 1) * LambNes.NUM_1_K), this.getPpuMemory().getPatternTable1(), 0x400, LambNes.NUM_1_K);
						}
						break;
					
					case MMC3Mapper.CONTROL_SWAP_2_1K_CHR_ROM_0800:
						logger.info("chr_rom 0800: address invert: " + this.isAddressInvert());
						if (!this.isAddressInvert())
						{
							// Select 2 KB CHR bank at PPU $0800-$0FFF
							System.arraycopy(this.getChrRom(), (value * LambNes.NUM_1_K), this.getPpuMemory().getPatternTable0(), 0x800, LambNes.NUM_1_K);
							System.arraycopy(this.getChrRom(), ((value + 1) * LambNes.NUM_1_K), this.getPpuMemory().getPatternTable0(), 0xC00, LambNes.NUM_1_K);
						}
						else
						{
							// Select 2 KB CHR bank at PPU $1800-$1FFF
							System.arraycopy(this.getChrRom(), (value * LambNes.NUM_1_K), this.getPpuMemory().getPatternTable1(), 0x800, LambNes.NUM_1_K);
							System.arraycopy(this.getChrRom(), ((value + 1) * LambNes.NUM_1_K), this.getPpuMemory().getPatternTable1(), 0xC00, LambNes.NUM_1_K);
						}
						break;
					
					case MMC3Mapper.CONTROL_SWAP_1K_CHR_ROM_1000:
						if (!this.isAddressInvert())
						{
							// Select 1 KB CHR bank at PPU $1000-$13FF
							System.arraycopy(this.getChrRom(), (value * LambNes.NUM_1_K), this.getPpuMemory().getPatternTable1(), 0, LambNes.NUM_1_K);
						}
						else
						{
							// Select 1 KB CHR bank at PPU $0000-$03FF
							logger.info("swapping 0x0000-0x03FF: " + value);
							System.arraycopy(this.getChrRom(), (value * LambNes.NUM_1_K), this.getPpuMemory().getPatternTable0(), 0, LambNes.NUM_1_K);
						}						
						break;
					
					case MMC3Mapper.CONTROL_SWAP_1K_CHR_ROM_1400:
						if (!this.isAddressInvert())
						{
							// Select 1 KB CHR bank at PPU $1400-$17FF
							System.arraycopy(this.getChrRom(), (value * LambNes.NUM_1_K), this.getPpuMemory().getPatternTable1(), 0x400, LambNes.NUM_1_K);
						}
						else
						{
							logger.info("swapping 0x0400-0x07FF: " + value);
							// Select 1 KB CHR bank at PPU $0400-$07FF
							System.arraycopy(this.getChrRom(), (value * LambNes.NUM_1_K), this.getPpuMemory().getPatternTable0(), 0x400, LambNes.NUM_1_K);
						}	
						break;
						
					case MMC3Mapper.CONTROL_SWAP_1K_CHR_ROM_1800: 
						if (!this.isAddressInvert())
						{
							// Select 1 KB CHR bank at PPU $1800-$1BFF
							System.arraycopy(this.getChrRom(), (value * LambNes.NUM_1_K), this.getPpuMemory().getPatternTable1(), 0x800, LambNes.NUM_1_K);
						}
						else
						{
							logger.info("swapping 0x0800-0x0BFF: " + value);
							// Select 1 KB CHR bank at PPU $0800-$0BFF
							System.arraycopy(this.getPrgRom(), (value * LambNes.NUM_1_K), this.getPpuMemory().getPatternTable0(), 0x800, LambNes.NUM_1_K);
						}	
						break;
						
					case MMC3Mapper.CONTROL_SWAP_1K_CHR_ROM_1C00:
						if (!this.isAddressInvert())
						{
							// Select 1 KB CHR bank at PPU $1C00-$1FFF
							System.arraycopy(this.getChrRom(), (value * LambNes.NUM_1_K), this.getPpuMemory().getPatternTable1(), 0xC00, LambNes.NUM_1_K);
						}
						else
						{
							logger.info("swapping 0x0C00-0x0FFF: " + value);
							// Select 1 KB CHR bank at PPU or $0C00-$0FFF
							System.arraycopy(this.getChrRom(), (value * LambNes.NUM_1_K), this.getPpuMemory().getPatternTable0(), 0xC00, LambNes.NUM_1_K);
						}	
						break;
						
					case MMC3Mapper.CONTROL_SWAP_8K_PRG_ROM_8000:
						logger.info("swapping out " + (this.prgRomBankMode == this.PRG_ROM_BANK_MODE_8000_9FFF_SWAPPABLE ? "8000" : "C000")  +  " with bank " + value);
						if (this.getPrgRomBankMode() == this.PRG_ROM_BANK_MODE_8000_9FFF_SWAPPABLE)
						{
							System.arraycopy(this.getPrgRom(), (value * LambNes.NUM_8_K), this.getCpuMemory().getMemory(), 0x8000, LambNes.NUM_8_K);
						}
						else
						{
							System.arraycopy(this.getPrgRom(), (value * LambNes.NUM_8_K), this.getCpuMemory().getMemory(), 0xC000, LambNes.NUM_8_K);
						}
						break;
						
					case MMC3Mapper.CONTROL_SWAP_8K_PRG_ROM_A000:
						logger.info("swapping out A000 with bank " + value);
						System.arraycopy(this.getPrgRom(), (value * LambNes.NUM_8_K), this.getCpuMemory().getMemory(), 0xA000, LambNes.NUM_8_K);
						break;
				}
			}
		}
		else if (address >= 0xA000 && address <= 0xBFFF)
		{
			if (addressEven)
			{
				logger.info("write to A000 register: " + value);
				// mirroring control
				this.setMirroringControl(value & 1);
			}
			else
			{
				logger.info("write to A001 register: " + value);
				// wram disable
				int wram = ((value & 128) >> 7);
				this.setPrgRamChipEnabled(wram ==1?false:true);
			}
		}
		else if (address >= 0xC000 && address <= 0xDFFF)
		{
			if (addressEven)
			{
				logger.info("write to C000 register: " + value);
				// IRQ counter reload.
			}
			else
			{
				logger.info("write to C001 register: " + value);
				// IRQ counter clear				
			}	
		}
		else if (address >= 0xE000 && address <= 0xFFFF)
		{
			if (addressEven)
			{
				logger.info("write to E000 register: " + value);
				// IRQ acknowledge
			}
			else
			{
				logger.info("write to E001 register: " + value);
				// IRQ enable				
			}
		}
		else
		{
			
		}
	}
	
	public int getPrgRomBankMode()
	{
		return prgRomBankMode;
	}

	public void setPrgRomBankMode(int prgRomBankMode)
	{
		this.prgRomBankMode = prgRomBankMode;
	}

	public boolean isPrgRamChipEnabled()
	{
		return prgRamChipEnabled;
	}

	public void setPrgRamChipEnabled(boolean prgRamChipEnabled)
	{
		this.prgRamChipEnabled = prgRamChipEnabled;
	}

	public int getMemoryFromHexAddress(int address)
	{
		int value = 0;
		if (address >= 0x6000 && address <= 0x7FFF && this.isPrgRamChipEnabled())
		{
			int prgRamIndex = address & 0x1FFF;
			value = this.getPrgRamValue(prgRamIndex);
		}
		else
		{
			value = super.getMemoryFromHexAddress(address);
		}
		
		return 	value;
	}
	
	public void setProgramInstructions(int[] programInstructions)
	{
		this.setPrgRom(programInstructions);
		
		//Object src, int srcPos, Object dest, int destPos, int length
		// load bank0 -- use first page from prgRom
		System.arraycopy(programInstructions, 0, this.getCpuMemory().getMemory(), NesCpuMemory.PRG_ROM_BASE, 16384);
		
		// load bank1 -- use last page from prgRom
		System.arraycopy(programInstructions, (programInstructions.length - LambNes.NUM_16_K), this.getCpuMemory().getMemory(), (NesCpuMemory.PRG_ROM_BASE + LambNes.NUM_16_K), LambNes.NUM_16_K);
	}

	public int[] getPrgRam()
	{
		return prgRam;
	}

	public void setPrgRam(int[] prgRam)
	{
		this.prgRam = prgRam;
	}
	
	public int getPrgRamValue(int prgRamIndex)
    {
    	return this.prgRam[prgRamIndex];
    }
	
	public void setPrgRamValue(int prgRamIndex, int value)
    {
    	this.prgRam[prgRamIndex] = value;
    }

	public int[] getPrgRom()
	{
		return prgRom;
	}

	public void setPrgRom(int[] prgRom)
	{
		this.prgRom = prgRom;
	}
	
	public boolean isAddressInvert()
	{
		return addressInvert;
	}

	public void setAddressInvert(boolean addressInvert)
	{
		this.addressInvert = addressInvert;
	}

	public int getPrgRomSwapControl()
	{
		return prgRomSwapControl;
	}

	public void setPrgRomSwapControl(int prgRomSwapControl)
	{
		this.prgRomSwapControl = prgRomSwapControl;
	}

	public int swapModeControl()
	{
		return swapModeControl;
	}

	public void setSwapModeControl(int chrRomSwapControl)
	{
		this.swapModeControl = chrRomSwapControl;
	}
	
    public int getMirroringControl()
	{
		return mirroringControl;
	}

	public void setMirroringControl(int mirroringControl)
	{
		logger.info("mirroring control: " + mirroringControl);
		this.mirroringControl = mirroringControl;
	}
	
	public int[] getChrRom()
	{
		return chrRom;
	}

	public void setChrRom(int[] chrRom)
	{
		this.chrRom = chrRom;
	}
}
