package com.lambelly.lambnes.platform.ppu;

import org.apache.commons.lang.ArrayUtils;
import org.apache.log4j.*;

import com.lambelly.lambnes.LambNes;
import com.lambelly.lambnes.platform.Platform;
import com.lambelly.lambnes.platform.ppu.registers.PPUControlRegister;
import com.lambelly.lambnes.util.BitUtils;
import com.lambelly.lambnes.util.NumberConversionUtils;
import java.awt.image.*;  

public class SpriteTile extends NesTile
{
	private int spriteNumber = 0;
	private int[] patternA = null;
	private int[] patternB = null;
	private SpriteAttribute spriteAttribute = new SpriteAttribute();
	private Logger logger = Logger.getLogger(SpriteTile.class);
	
	public SpriteTile(int spriteNumber)
	{
		this.setSpriteNumber(spriteNumber);
		this.instantiateTile();
	}
	
	public SpriteTile(int spriteNumber, SpriteAttribute attribute)
	{
		this.setSpriteNumber(spriteNumber);
		this.setAttributes(attribute);
		this.instantiateTile();
	}
	
	public SpriteTile (SpriteTile tile)
	{
		this.setSpriteNumber(tile.getSpriteNumber());
		this.setPatternA(tile.getPatternA());
		this.setPatternB(tile.getPatternB());
	}
	
	public int[] getTileByteArray()
	{
		int spriteLength;
		int address;
		
		// determine length and address -- 8x16 logic
		if (LambNes.getPlatform().getPpu().getPpuControlRegister().getSpriteSize() == PPUControlRegister.SPRITE_SIZE_8X8)
		{
			int tileNumber = this.getAttributes().getTileIndex();
			logger.info("not using 8x16 logics");
			address = (this.getNameTableSelectBit() * 0x1000) | (tileNumber * 16);
			spriteLength = 16;
		}
		else
		{
			logger.info("using 8x16 logics");
			logger.info("8x16: tile: " + this.getAttributes().getTileIndex());
			address = ((this.getAttributes().getTileIndex() & 1) == 1?0x1000:0 ) + ((this.getAttributes().getTileIndex() & 0xFE) * 16);
			logger.info("tileNumber: " + this.getAttributes().getTileIndex() + " address: " + address);
			spriteLength = 32;
		}
		
		int[] tile = new int[spriteLength];
		
		for (int x = 0; x < spriteLength; x++)
		{
			int value = LambNes.getPlatform().getPpuMemory().getMemoryFromHexAddress(address + x);
		
			tile[x] = value;
		}		
		
		return tile;
	}	
	
	protected void instantiateTile()
	{
		// logger.debug("instantiating sprite");
		
		int[] sprite = this.getTileByteArray();			
		
		// for (int i = 0; i < sprite.length; i++)
		// {
		// 		logger.debug("sprite array for sprite " + this.getSpriteNumber() + ": row: " + i + ": " + sprite[i]);
		// }
		
		if (LambNes.getPlatform().getPpu().getPpuControlRegister().getSpriteSize() == PPUControlRegister.SPRITE_SIZE_8X8)
		{
			this.setPatternA(ArrayUtils.subarray(sprite, 0, 8));
			this.setPatternB(ArrayUtils.subarray(sprite, 8, 16));
		}
		else
		{
			this.setPatternA(new int[16]);
			this.setPatternB(new int[16]);

			logger.info("sprite size: " + sprite.length);
			
			System.arraycopy(sprite, 0, this.getPatternA(), 0, 8);
			System.arraycopy(sprite, 8, this.getPatternB(), 0, 8);
			System.arraycopy(sprite, 16, this.getPatternA(), 8, 8);
			System.arraycopy(sprite, 24, this.getPatternB(), 8, 8);
		}
	}
	
	public PaletteColor[] getTileColorRow(int row)
	{
		PaletteColor colorBit[] = new PaletteColor[8];
		for (int i = 0; i < 8; i++)
		{
			PaletteColor pixel = super.getPpuMemory().getSpritePaletteColor(this.getPixelColorPaletteIndex(i, row));
			colorBit[i ^ 0x07] = pixel;
			logger.info("setting colorBit " + (i ^ 0x7) +" to mpc: " + pixel.getMasterPaletteIndex() + " pi: " + pixel.getPaletteIndex() + " for row: " + row + " of sprite: " + this.getSpriteNumber());
		}
		
		return colorBit;
	}
	
	public int getPixelColorPaletteIndex(int column, int row)
	{
		// staleness check -- occurs sometimes
		if (this.getPatternA() == null) 
		{
			//logger.info("refreshing tile");
			this.refreshTile();
		}
		
		// logger.debug("generating pixel color for sprite: " + this.getAttributes().getTileIndex() + ", spriteNumber: " + this.getSpriteNumber() + " column:" + column + ", row: " + row + ", msb: " + this.getAttributes().getColorHighBit());
		// logger.debug("pattern a row " + row + ": " + this.getPatternA()[row]);
		// logger.debug("pattern b row " + row + ": " + this.getPatternB()[row]);
		
		// horizontal flip
		if (this.getAttributes().isHorizontalFlip())
		{
			// flip column's bits
			column ^= 0x07;
		}
		
		// vertical flip
		if (this.getAttributes().isVerticalFlip())
		{
			// flip row's bits
			row ^= 0x07;
		}
		
		int color = 0;
		color = (this.getAttributes().getColorHighBit() << 2) | 
			(((BitUtils.isBitSet(this.getPatternB()[row],column))?1:0) << 1) | 
			((BitUtils.isBitSet(this.getPatternA()[row],column))?1:0);
 
		logger.debug("pattern a pixel: " + ((BitUtils.isBitSet(this.getPatternA()[row],column))?1:0));
		logger.debug("pattern b pixel: " + (((BitUtils.isBitSet(this.getPatternB()[row],column))?1:0) << 1));
		logger.debug("color bitstring generated for sprite " + this.getSpriteNumber()+ ": c: " + column + ", r: " + row + ": " + Integer.toBinaryString(color));
		return color;
	}
	
	public void refreshAttributes()
	{
		this.setAttributes(LambNes.getPlatform().getPpuMemory().getSprRamForSpriteNumber(this.getSpriteNumber()));
	}
	
	@Override
	public int getNameTableSelectBit()
	{
		return LambNes.getPlatform().getPpu().getPpuControlRegister().getSpritePatternTableAddress();
	}
	
	public String toString()
	{
		int rows = (this.patternA.length);
		
		// generate color map string
		String colorMapString = "";
		String patternA = "patternA: \n";
		String patternB = "patternB: \n";
		
		for (int row = 0; row < rows; row++)
		{
			patternA += "\t" + NumberConversionUtils.generateBinaryStringWithleadingZeros(this.patternA[row], 8) + "\n";
			patternB += "\t" + NumberConversionUtils.generateBinaryStringWithleadingZeros(this.patternB[row], 8) + "\n";
			colorMapString+= "\n";
			for (int col = 0; col < 8; col++)
			{
				colorMapString += this.getPixelColorPaletteIndex(col, row);
			}
		}
		
		return patternA + 
			patternB +
			"\t\tcolor high bit: " + this.getAttributes().getColorHighBit() + "\n" +
			"\t\tcolor map: " + colorMapString;
	}

	public int getSpriteNumber()
	{
		return spriteNumber;
	}

	public void setSpriteNumber(int spriteNumber)
	{
		this.spriteNumber = spriteNumber;
	}

	public int[] getPatternA()
	{
		return patternA;
	}

	public void setPatternA(int[] patternA)
	{
		this.patternA = patternA;
	}

	public int[] getPatternB()
	{
		return patternB;
	}

	public void setPatternB(int[] patternB)
	{
		this.patternB = patternB;
	}

	public SpriteAttribute getAttributes()
	{
		return spriteAttribute;
	}
	
	public void setAttributes(int spriteNumber)
	{
		this.spriteAttribute = LambNes.getPlatform().getPpuMemory().getSprRamForSpriteNumber(spriteNumber);
	}
	
	public void setAttributes(NesTileAttribute spriteAttributes)
	{
		this.spriteAttribute = (SpriteAttribute)spriteAttributes;
	}
}
